#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting or consulting/
#    support services related to the Software), a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)

# net class definition.
from PySide6.QtCore import (QPoint, Qt, QLineF, QRectF, QPointF, )
from PySide6.QtGui import (QPen, QStaticText, QPainterPath, QFont, )
from PySide6.QtWidgets import (QGraphicsLineItem, QGraphicsItem, QGraphicsPathItem,
                               QGraphicsEllipseItem, QGraphicsSceneMouseEvent,
                               QGraphicsSceneHoverEvent, )
import pdk.schLayers as schlyr
import math
import itertools as itt
from typing import Union, Optional, NamedTuple


class crossingDot(QGraphicsEllipseItem):
    dotDiameter = schlyr.crossingDotDiameter

    def __init__(self, point: QPoint):

        self.point = point
        self._radius = crossingDot.dotDiameter
        super().__init__(point.x() - self._radius, point.y() - self._radius,
                         2 * self._radius, 2 * self._radius)
        self.setPen(schlyr.wirePen)
        self.setBrush(schlyr.wireBrush)
        self._name = None

    def paint(self, painter, option, widget) -> None:
        if self.isSelected():
            painter.setPen(schlyr.selectedWirePen)
            painter.setBrush(schlyr.selectedWireBrush)
        else:
            painter.setPen(schlyr.wirePen)
            painter.setBrush(schlyr.wireBrush)
        painter.drawEllipse(self.point, self._radius, self._radius)

    def __repr__(self):
        return f"crossingDot({self.point},{self._radius})"

    def findNets(self) -> set["schematicNet"]:
        if self.scene():
            return {netItem for netItem in self.scene().items(self.sceneBoundingRect()) if
                    isinstance(netItem, schematicNet)}
        else:
            return set()


class selfIndNetIndTuple(NamedTuple):
    selfIndex: int
    net: QGraphicsItem
    netEndIndex: int


class pointSelfIndex(NamedTuple):
    point: QPoint
    selfIndex: int


class schematicNet(QGraphicsItem):
    def __init__(self, start: QPoint, end: QPoint, mode: int = 0):
        super().__init__()
        self.setFlag(QGraphicsItem.ItemIsMovable, True)
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)
        self.setFlag(QGraphicsItem.ItemIsFocusable, True)
        # self.setFlag(QGraphicsItem.ItemSendsScenePositionChanges, True)
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges, True)
        self._mode = 0
        self._name: str = ""
        self._nameConflict: bool = False
        self._nameAdded: bool = False
        self._nameSet: bool = False
        self._highlighted: bool = False
        self._flightLinesSet: set[schematicNet] = set()
        self._connectedNetsSet: set[schematicNet] = set()
        self._netIndexTupleSet: set[selfIndNetIndTuple] = set()
        self._pinLocIndexSet: set[pointSelfIndex] = set()
        self._pinSnapLines: dict[int, set[guideLine]] = {}
        self._netSnapLines: dict[int, set[guideLine]] = {}
        self._stretch: bool = False
        self._nameFont = QFont()
        self._nameFont.setPointSize(8)
        self._nameItem = QStaticText(self._name)
        self._draftLine = QLineF(start, end)
        self._stretchLine: Optional[guideLine] = None
        match self._mode:
            case 0:
                self._angle = 90 * math.floor(((self._draftLine.angle() + 45) % 360) / 90)
            case 1:
                self._angle = 45 * math.floor(((self._draftLine.angle() + 22.5) % 360) / 45)
        self._draftLine.setAngle(0)
        self.setTransformOriginPoint(self._draftLine.p1())
        if self.scene():
            self._draftLine.setP2(
                self.scene().snapToGrid(self._draftLine.p2(), self.scene().snapTuple))
        self._shapeRect = QRectF(self._draftLine.p1(), self._draftLine.p2()).adjusted(-2,
                                                                                      -2, 2,
                                                                                      2)
        self._boundingRect = self._shapeRect.adjusted(-8, -8, 8, 8)
        self.setRotation(-self._angle)

    @property
    def draftLine(self):
        return self._draftLine

    @draftLine.setter
    def draftLine(self, line: QLineF):
        self.prepareGeometryChange()
        self._draftLine = line
        match self._mode:
            case 0:
                self._angle = 90 * math.floor(((self._draftLine.angle() + 45) % 360) / 90)
            case 1:
                self._angle = 45 * math.floor(((self._draftLine.angle() + 22.5) % 360) / 45)
        self._draftLine.setAngle(0)
        self.setTransformOriginPoint(self._draftLine.p1())
        if self.scene():
            self._draftLine.setP2(
                self.scene().snapToGrid(self._draftLine.p2(), self.scene().snapTuple))
        self._shapeRect = (
            QRectF(self._draftLine.p1(), self._draftLine.p2()).normalized().adjusted(-2, -2,
                                                                                     2, 2))
        self._boundingRect = self._shapeRect.adjusted(-8, -8, 8, 8)
        self.setRotation(-self._angle)

    def shape(self) -> QPainterPath:
        path = QPainterPath()
        path.addRect(self._shapeRect)
        return path

    def boundingRect(self):
        return self._boundingRect

    def paint(self, painter, option, widget=...):
        painter.setFont(self._nameFont)
        if self.isSelected():
            pen = schlyr.selectedWirePen
            if self._stretch:
                pen = schlyr.stretchWirePen
        elif self._highlighted:
            pen = schlyr.hilightPen
        else:
            pen = schlyr.wirePen

        painter.setPen(pen)
        painter.drawLine(self._draftLine)
        if self._nameSet:
            painter.save()
            painter.translate(self._draftLine.center().x(), self._draftLine.center().y())
            painter.rotate(self._angle)
            painter.drawStaticText(0, 0, self._nameItem)
            painter.restore()

    def sceneEvent(self, event):
        """
        Handle events related to the scene.

        Args:
            event: The event to handle.

        Returns:
            True if the event was handled successfully, False otherwise.
        """
        # Check if the current scene has the drawWire edit mode enabled
        if self.scene() and self.scene().editModes.drawWire:
            return False
        else:
            # Call the parent class's sceneEvent method to handle the event
            super().sceneEvent(event)
        return True

    def __repr__(self):
        return f"schematicNet({self.sceneEndPoints})"

    def itemChange(self, change, value):
        if change == QGraphicsItem.ItemSceneChange and not value:
            self.mergeOrthoNets()
            self.clearDots()
        return super().itemChange(change, value)

    def mergeOrthoNets(self):
        cdots = self.findDots()
        for dot in cdots:
            orthoNets = list(filter(self.isOrthogonal, dot.findNets()))
            if orthoNets:
                self.scene().mergeNets(orthoNets[0])

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent):
        self.mergeOrthoNets()
        self.clearDots()

        if self._stretch:
            self.startStretch(event)
        else:
            self.findSymPinConnections()
            self.findNetConnections()
            self.createPinSnapLines()
            self.createNetSnapLines()
        super().mousePressEvent(event)

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:

        if self.stretch:
            self.extendStretch(event)
        else:
            if self._pinSnapLines:
                self.extendPinSnapLines()
            if self._netSnapLines:
                self.extendNetSnapLines()
        super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent):

        if self.stretch:
            self.endStretch()
        else:
            self.removePinSnapLines()
            self.removeNetSnapLines()
        # self.createDots()
        super().mouseReleaseEvent(event)

    def createPinSnapLines(self):
        for tupleItem in self._pinLocIndexSet:
            lineSet = self._pinSnapLines.setdefault(tupleItem.selfIndex, set())
            stretchLine = guideLine(tupleItem.point, tupleItem.point)
            lineSet.add(stretchLine)

    def createNetSnapLines(self):
        for tupleItem in self._netStretchTupleSet:
            print(f'net name is set: {tupleItem.net.nameSet}')
            lineSet = self._netSnapLines.setdefault(tupleItem.selfIndex, set())
            stretchLine = guideLine(
                tupleItem.net.sceneEndPoints[tupleItem.netEndIndex - 1],
                self.sceneEndPoints[tupleItem.selfIndex])
            if tupleItem.net.nameSet:
                stretchLine.name = tupleItem.net.name
                stretchLine.nameSet = True
            lineSet.add(stretchLine)
            if tupleItem.net.scene():
                self.scene().removeItem(tupleItem.net)
                self.scene().addItem(stretchLine)

    def extendPinSnapLines(self):
        for index, lineSet in self._pinSnapLines.items():
            for snapLine in lineSet:
                if not snapLine.scene():
                    self.scene().addItem(snapLine)
                snapLine.setLine(QLineF(snapLine.mapFromScene(self.sceneEndPoints[index]),
                                        snapLine.line().p2()))

    def extendNetSnapLines(self):
        for index, lineSet in self._netSnapLines.items():
            for snapLine in lineSet:
                snapLine.setLine(QLineF(snapLine.line().p1(),
                                        snapLine.mapFromScene(self.sceneEndPoints[index])))

    def removePinSnapLines(self):
        for snapLineSet in self._pinSnapLines.values():
            lines = []
            for snapLine in snapLineSet:
                lines = self.scene().addStretchWires(
                    snapLine.mapToScene(snapLine.line().p1()).toPoint(),
                    snapLine.mapToScene(snapLine.line().p2()).toPoint(), )
                if snapLine.scene():
                    self.scene().removeItem(snapLine)
                if lines:
                    self.scene().addListUndoStack(lines)

        self._pinSnapLines = dict()

    def removeNetSnapLines(self):
        for snapLineSet in self._netSnapLines.values():

            for snapLine in snapLineSet:
                lines = self.scene().addStretchWires(
                    snapLine.mapToScene(snapLine.line().p1()).toPoint(),
                    snapLine.mapToScene(snapLine.line().p2()).toPoint())
                if lines:
                    self.scene().addListUndoStack(lines)
                    for line in lines:
                        if snapLine.nameSet:
                            line.name = snapLine.name
                            line.nameSet = True
                if snapLine.scene():
                    self.scene().removeItem(snapLine)

        self._netSnapLines = dict()

    def startStretch(self, event):
        """
        Handle the start of the stretch of the net using one of the end points.
        """

        eventPos = event.scenePos().toPoint()
        self.setFlag(QGraphicsItem.ItemIsMovable, False)
        if (eventPos - self.mapToScene(
                self._draftLine.p1()).toPoint()).manhattanLength() <= self.scene().snapDistance:
            self.setCursor(Qt.SizeHorCursor)
            self._stretchLine = guideLine(self.mapToScene(self._draftLine.p1()), eventPos)
        elif (eventPos - self.mapToScene(
                self._draftLine.p2()).toPoint()).manhattanLength() <= self.scene().snapDistance:
            self.setCursor(Qt.SizeHorCursor)
            self._stretchLine = guideLine(self.mapToScene(self._draftLine.p2()), eventPos)

    def extendStretch(self, event):
        eventPos = event.scenePos().toPoint()
        if self._stretchLine is not None:
            if self._stretchLine.scene() is None:
                self.scene().addItem(self._stretchLine)
            self._stretchLine.setLine(QLineF(self._stretchLine.line().p1(),
                                             self._stretchLine.mapFromScene(eventPos), ))

    def endStretch(self):
        self._stretch = False
        self._stretchSide = None
        if self._stretchLine and self.scene():
            lines = self.scene().addStretchWires(*self._stretchLine.sceneEndPoints)
            self.scene().removeItem(self._stretchLine)
        if lines:
            self.scene().addListUndoStack(lines)
            self.scene().mergeSplitNets(lines[0])
        self._stretchLine = None
        self.setCursor(Qt.ArrowCursor)

    @property
    def innerRect(self) -> QRectF:
        return (
            QRectF(self._draftLine.p1(), self._draftLine.p2()).normalized().adjusted(2, 2,
                                                                                     -2,
                                                                                     -2))

    @property
    def sceneInnerRect(self) -> QRectF:
        return self.mapRectToScene(self.innerRect)

    def hoverEnterEvent(self, event: QGraphicsSceneHoverEvent) -> None:
        """
        Override the hoverEnterEvent method of QGraphicsItem.

        Args:
            event (QGraphicsSceneHoverEvent): The hover event.

        Returns:
            None
        """
        super().hoverEnterEvent(event)

        # Check if highlightNets flag is set in the scene
        if self.scene().highlightNets:
            # Create a set of connected netItems based on certain conditions
            self._connectedNetsSet = {netItem for netItem in self.scene().items() if (
                    isinstance(netItem, schematicNet) and (
                    self.nameSet or self.nameAdded) and netItem.name == self.name)}

            # Highlight the connected netItems
            for netItem in self._connectedNetsSet:
                netItem.highlight()

            # Create flight lines and add them to the scene
            for netItem in self._connectedNetsSet:
                flightLine = netFlightLine(self.mapToScene(self._draftLine.center()),
                                           netItem.mapToScene(netItem.draftLine.center()), )
                self._flightLinesSet.add(flightLine)
                self.scene().addItem(flightLine)

    def hoverLeaveEvent(self, event: QGraphicsSceneHoverEvent) -> None:
        super().hoverLeaveEvent(event)
        if self._highlighted:
            for flightLine in self._flightLinesSet:
                self.scene().removeItem(flightLine)
            self._flightLinesSet = set()
        [netItem.unhighlight() for netItem in self._connectedNetsSet]

    def isParallel(self, otherNet: "schematicNet") -> bool:
        return abs((self.angle - otherNet.angle) % 180) < 1

    def isOrthogonal(self, otherNet: "schematicNet") -> bool:
        return abs((self.angle - otherNet.angle - 90) % 180) < 1

    def notParallel(self, otherNet: "schematicNet") -> bool:
        return not self.isParallel(otherNet)

    def containsDot(self, dot: crossingDot) -> bool:
        """
        Check if the dot is contained within the scene shape rectangle.

        Args:
            dot: The dot to check.

        Returns:
            True if the dot is contained within the scene shape rectangle, False otherwise.
        """
        if self.sceneShapeRect.contains(self.mapFromScene(dot.point)):
            return True
        return False

    def findOverlapNets(self) -> set["schematicNet"]:
        """
        Find all netItems in the scene that overlap with self.sceneShapeRect.

        Returns:
            set: A set of netItems that overlap with self.sceneShapeRect.
        """
        if self.scene():
            overlapNets = {netItem for netItem in self.scene().items(self.sceneShapeRect) if
                           isinstance(netItem, schematicNet) and netItem is not self}
        else:
            overlapNets = set()
        return overlapNets

    def findCrossingNets(self, otherNets: set) -> dict[int, list["schematicNet"]]:
        """
        Finds the crossing nets in the scene.

        Returns:
            dict: A dictionary with the index of each scene end point as the key, and a list
            of nets that cross that end point as the value.
        """

        orthoNets = list(filter(self.isOrthogonal, otherNets))
        crossingNets = dict()
        for netEnd in self.sceneEndPoints:
            crossingNets[self.sceneEndPoints.index(netEnd)] = list()
            for netItem in orthoNets:
                if netItem.sceneInnerRect.contains(netEnd):
                    crossingNets[self.sceneEndPoints.index(netEnd)].append(netItem)
        return crossingNets

    def createSplitNets(self, crossingNets: dict) -> set:
        """
        Create split nets based on the net connections.

        Args:
            crossingNets (dict): A dictionary containing the net connections. Keys are
            netEnds. Values are lists of netItems that are connected to the netEnd.

        Returns:
            set: A set of tuples representing netEnd they are connected,
            themselves and their connected end.
        """
        splitNetTuples = set()

        for endIndex, orthoNets in crossingNets.items():
            for orthoNet in orthoNets:
                # Create new net from endIndex to orthoNet draftLine.p2()
                newNet1 = schematicNet(self.sceneEndPoints[endIndex],
                                       orthoNet.mapToScene(orthoNet.draftLine.p2()), )

                # Create new net from orthoNet draftLine.p1() to endIndex
                newNet2 = schematicNet(orthoNet.mapToScene(orthoNet.draftLine.p1()),
                                       self.sceneEndPoints[endIndex], )
                # Set name and nameSet properties of newNet1 and newNet2
                if self._nameSet or self._nameAdded:
                    newNet1.name = self._name
                    newNet2.name = self._name
                    newNet1.nameSet = self._nameSet
                    newNet2.nameSet = self._nameSet
                    newNet1.nameAdded = self._nameAdded
                    newNet2.nameAdded = self._nameAdded
                self.scene().removeItem(orthoNet)
                splitNetTuples.add(selfIndNetIndTuple(endIndex, newNet1, 0))
                splitNetTuples.add(selfIndNetIndTuple(endIndex, newNet2, 1))
        return splitNetTuples

    def mergeNets(self) -> tuple["schematicNet", "schematicNet"]:
        """
        Merges overlapping nets and returns the merged net.

        Returns:
            Optional[schematicNet]: The merged net if there are overlapping nets, otherwise returns self.
        """
        # Find other nets that overlap with self
        otherNets = self.findOverlapNets()

        # If there are other nets
        if otherNets:
            # Filter the other nets to find the parallel ones
            parallelNets = [netItem for netItem in otherNets if self.isParallel(netItem)]

            # If there are parallel nets
            if parallelNets:
                # Create an initialRect variable and set it to self's sceneShapeRect
                initialRect = self.sceneShapeRect

                # Iterate over the parallel nets
                for netItem in parallelNets:
                    # Update the initialRect by uniting it with each parallel net's
                    # sceneShapeRect
                    initialRect = initialRect.united(netItem.sceneShapeRect)

                    # Remove each parallel net from the scene
                    if netItem.scene():
                        netItem.scene().removeItem(netItem)

                # Adjust the initialRect by 2 pixels on each side
                newNetPoints = initialRect.adjusted(2, 2, -2, -2)

                # Get the coordinates of the adjusted rectangle
                x1, y1, x2, y2 = newNetPoints.getCoords()

                # Create a new schematicNet with the snapped coordinates
                newNet = schematicNet(self.snapToGrid(QPoint(x1, y1)),
                                      self.snapToGrid(QPoint(x2, y2)))

                # If self has a name, set the name and nameSet of the newNet
                if self._nameSet or self._nameAdded:
                    newNet.name = self._name
                    newNet.nameSet = self._nameSet
                    newNet.nameAdded = self._nameAdded

                return self, newNet  # original net, new net

        # If there are no other nets or no parallel nets, return self
        return self, self

    def clearDots(self):
        """
        Clears all crossingDot items from the net's scene shape area.
        """
        if self.scene():
            crossing_dots = [item for item in self.scene().items(self.sceneShapeRect) if
                             isinstance(item, crossingDot)]
            for dot in crossing_dots:
                self.scene().removeItem(dot)

    def createDots(self):
        """
        Create crossing dots at the intersection of nets.

        This function finds overlapping nets, generates combinations of two nets,
        and checks if they intersect at the same endpoints. If they do, a crossing
        dot is created at that intersection point.

        Parameters:
        - self: The instance of the class.

        Returns:
        - None
        """
        # Find overlapping nets
        otherNets = self.findOverlapNets()

        # Generate combinations of two nets
        netCombinations = set(itt.combinations(otherNets, 2))

        # Iterate over net combinations
        for netCombination in netCombinations:
            net2, net3 = netCombination

            # Check if both nets exist
            if net2 and net3:
                # Iterate over all possible combinations of endpoints
                for netEnd1, netEnd2, netEnd3 in itt.product(self.sceneEndPoints,
                                                             net2.sceneEndPoints,
                                                             net3.sceneEndPoints):
                    # Check if all endpoints are the same
                    if netEnd1 == netEnd2 and netEnd2 == netEnd3:
                        # Create a crossing dot at the intersection point if there is no dot
                        # there.
                        dots = [dotItem for dotItem in self.scene().items(QRectF(netEnd1,
                                                                                 netEnd1).adjusted(
                            -2, -2, 2, 2)) if isinstance(dotItem,
                                                         crossingDot)]
                        if not dots:
                            newDot = crossingDot(netEnd1)
                            self.scene().addItem(newDot)

    def findDots(self) -> list[crossingDot]:
        """
        Find all crossingDot items in the net shape.

        Returns:
            set: A set of crossingDot items.
        """
        crossing_dots = []
        if self.scene():
            crossing_dots = [item for item in self.scene().items(self.sceneShapeRect) if
                             isinstance(item, crossingDot)]
        return crossing_dots

    def findSymPinConnections(self):
        """
        Find the pin connections to a symbol in the scene.
        """

        # Set to store the index of pin locations
        self._pinLocIndexSet = set()
        connectedPins = []
        if self.scene():
            # Find the connected pins in the scene
            connectedPins = self.scene().findRectSymbolPin(self.sceneShapeRect)
        if connectedPins:
            # Iterate over the end points and connected pins
            for end, pin in itt.product(self.sceneEndPoints, connectedPins):
                # Check if the pin's bounding rectangle contains the end point
                if pin.sceneBoundingRect().contains(end):
                    # Add the index of the end point to the set
                    self._pinLocIndexSet.add(
                        pointSelfIndex(end, self.sceneEndPoints.index(end)))

    def findNetConnections(self):
        # Find the overlap nets
        overlapNets = self.findOverlapNets()
        self._netStretchTupleSet = set()
        for netItem in overlapNets:
            for selfEnd, netItemEnd in itt.product(self.sceneEndPoints,
                                                   netItem.sceneEndPoints):
                if selfEnd == netItemEnd:
                    self._netStretchTupleSet.add(
                        selfIndNetIndTuple(self.sceneEndPoints.index(selfEnd), netItem,
                                           netItem.sceneEndPoints.index(netItemEnd)))

    def snapToGrid(self, point: Union[QPoint, QPointF]) -> QPoint:
        if self.scene():
            return self.scene().snapToGrid(point, self.scene().snapTuple)
        else:
            return point

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        if name != "":  # net name should not be an empty string
            self.prepareGeometryChange()
            self._name = name
            self._nameItem = QStaticText(self._name)

    @property
    def nameSet(self) -> bool:
        """
        Check if the name of the net is explicitly set.

        Returns:
            bool: The value of the 'nameSet' attribute.
        """
        return self._nameSet

    @nameSet.setter
    def nameSet(self, value: bool):
        """
        If the name of the net is explicitly set, set this attribute to True.

        Args:
            value (bool): The value to set for the 'nameSet' attribute.

        Raises:
            AssertionError: If the input value is not a boolean.
        """
        assert isinstance(value, bool)
        self._nameSet = value

    @property
    def nameAdded(self) -> bool:
        """
        Name added is true if net name is set due to a connected net or pin.
        """
        return self._nameAdded

    @nameAdded.setter
    def nameAdded(self, value: bool):
        assert isinstance(value, bool)
        self._nameAdded = value

    @property
    def nameConflict(self) -> bool:
        """
        If two different names are attempted to be set for the net.
        """
        return self._nameConflict

    @nameConflict.setter
    def nameConflict(self, value: bool):
        assert isinstance(value, bool)
        self._nameConflict = value

    @property
    def sceneEndPoints(self):
        return [self.mapToScene(self._draftLine.p1()).toPoint(),
                self.mapToScene(self._draftLine.p2()).toPoint(), ]

    def highlight(self):
        self._highlighted = True
        self.update()

    def unhighlight(self):
        self._highlighted = False
        self.update()

    @property
    def highlighted(self):
        return self._highlighted

    @highlighted.setter
    def highlighted(self, value: bool):
        self._highlighted = value

    @property
    def sceneShapeRect(self) -> QRectF:
        return self.mapRectToScene(self._shapeRect).normalized().toRect()

    @property
    def stretch(self) -> bool:
        return self._stretch

    @stretch.setter
    def stretch(self, value: bool):
        self._stretch = value

    @property
    def stretchSide(self):
        """
        The end where the net is stretched.
        """
        return self._stretchSide

    @stretchSide.setter
    def stretchSide(self, value: str):
        self.prepareGeometryChange()
        self._stretchSide = value

    @property
    def angle(self):
        return self._angle


class netFlightLine(QGraphicsPathItem):
    wireHighlightPen = QPen(schlyr.wireHilightLayer.pcolor, schlyr.wireHilightLayer.pwidth,
                            schlyr.wireHilightLayer.pstyle, )

    def __init__(self, start: QPoint, end: QPoint):
        self._start = start
        self._end = end
        super().__init__()

    def __repr__(self):
        return f"netFlightLine({self.mapToScene(self._start)},{self.mapToScene(self._end)})"

    def paint(self, painter, option, widget) -> None:
        painter.setPen(netFlightLine.wireHighlightPen)
        line = QLineF(self._start, self._end)
        perpendicularLine = QLineF(line.center(),
                                   line.center() + QPointF(-line.dy(), line.dx()))
        perpendicularLine.setLength(100)

        path = QPainterPath()
        path.moveTo(self._start)
        path.quadTo(perpendicularLine.p2(), self._end)
        painter.drawPath(path)


class guideLine(QGraphicsLineItem):
    def __init__(self, start: QPoint, end: QPoint):
        self._start = start
        self._end = end
        super().__init__(QLineF(self._start, self._end))
        self.setPen(schlyr.guideLinePen)
        self._name: str = ""
        self._nameSet: bool = False

    @property
    def sceneEndPoints(self) -> list[QPoint]:
        return [self.mapToScene(self.line().p1()).toPoint(),
                self.mapToScene(self.line().p2()).toPoint(), ]

    def __repr__(self):
        return f"guideLine({self.mapToScene(self.line().p1()), self.mapToScene(self.line().p2())}"

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        if name != "":  # net name should not be an empty string
            self._name = name

    @property
    def nameSet(self) -> bool:
        return self._nameSet

    @nameSet.setter
    def nameSet(self, value: bool):
        assert isinstance(value, bool)
        self._nameSet = value



#   “Commons Clause” License Condition v1.0
#  #
#   The Software is provided to you by the Licensor under the License, as defined
#   below, subject to the following condition.
#  #
#   Without limiting other conditions in the License, the grant of rights under the
#   License will not include, and the License does not grant to you, the right to
#   Sell the Software.
#  #
#   For purposes of the foregoing, “Sell” means practicing any or all of the rights
#   granted to you under the License to provide to third parties, for a fee or other
#   consideration (including without limitation fees for hosting or consulting/
#   support services related to the Software), a product or service whose value
#   derives, entirely or substantially, from the functionality of the Software. Any
#   license notice or attribution required by the License must also include this
#   Commons Clause License Condition notice.
#  #
#   Software: Revolution EDA
#   License: Mozilla Public License 2.0
#   Licensor: Revolution Semiconductor (Registered in the Netherlands)

from PySide6.QtCore import (Qt)
from PySide6.QtGui import (QPen, QColor, QColorConstants)

from revedaEditor.common.layers import (wireLayer, symbolLayer,
                                        selectedWireLayer, pinLayer,
                                        labelLayer,
                                        textLayer, draftLayer)


class sPen(QPen):
    def __init__(self, pname: str, pcolor: QColor, pwidth: int,
                 plinestyle: Qt.PenStyle = Qt.SolidLine):
        super().__init__()
        self._pname = pname
        self._pcolor = pcolor
        self._pwidth = pwidth
        self._plinestyle = plinestyle
        self.setWidth(self._pwidth)
        self.setColor(self._pcolor)
        self.setStyle(self._plinestyle)

    @property
    def pname(self):
        return self._pname

    @classmethod
    def returnPen(cls, penName):
        match penName:
            case 'wirePen':
                rpen = cls('wirePen', wireLayer.color, 2)
                rpen.setCosmetic(True)
            case 'symbolPen':
                rpen = cls("symbolPen", symbolLayer.color, 2)
                rpen.setCosmetic(True)
            case 'selectedWirePen':
                rpen = cls("selectedWirePen", selectedWireLayer.color, 2)
            case 'pinPen':
                rpen = cls('pinPen', pinLayer.color, 2)
            case 'labelPen':
                rpen = cls('labelPen', labelLayer.color, 1)
            case 'textPen':
                rpen = cls('textPen', textLayer.color, 1)
            case 'draftPen':
                rpen = cls('draftPen', draftLayer.color, 2, Qt.DashLine)
            case 'hilightPen':
                rpen = cls('hilightPen', QColorConstants.DarkMagenta, 5, Qt.SolidLine)
            case other:
                rpen = cls('otherPen', QColor('darkGray'), 2, Qt.DotLine)
        return rpen



#   “Commons Clause” License Condition v1.0
#  #
#   The Software is provided to you by the Licensor under the License, as defined
#   below, subject to the following condition.
#  #
#   Without limiting other conditions in the License, the grant of rights under the
#   License will not include, and the License does not grant to you, the right to
#   Sell the Software.
#  #
#   For purposes of the foregoing, “Sell” means practicing any or all of the rights
#   granted to you under the License to provide to third parties, for a fee or other
#   consideration (including without limitation fees for hosting or consulting/
#   support services related to the Software), a product or service whose value
#   derives, entirely or substantially, from the functionality of the Software. Any
#   license notice or attribution required by the License must also include this
#   Commons Clause License Condition notice.
#  #
#   Software: Revolution EDA
#   License: Mozilla Public License 2.0
#   Licensor: Revolution Semiconductor (Registered in the Netherlands)

import math

# shape class definition for symbol editor.
# base class for all shapes: rectangle, circle, line
from PySide6.QtCore import QPoint, QRect, QRectF, Qt, QLine, QLineF
from PySide6.QtGui import (QPen, QFont, QFontMetrics, QPainterPath, QTextOption,
                           QFontDatabase, )
from PySide6.QtWidgets import (QGraphicsItem, QGraphicsSceneMouseEvent, )
from quantiphy import Quantity
import revedaEditor.common.net as net
import revedaEditor.backend.dataDefinitions as ddef
import pdk.callbacks as cb


class shape(QGraphicsItem):
    def __init__(self, pen: QPen, gridTuple: tuple) -> None:
        super().__init__()
        self.setFlag(QGraphicsItem.ItemIsMovable, True)
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges, True)
        self.setFlag(QGraphicsItem.ItemIsFocusable, True)
        self.setAcceptHoverEvents(True)
        # self.setZValue(self.layer.z)
        self._pen = pen
        self._gridTuple = gridTuple
        self._angle = 0  # rotation angle
        self._stretch: bool = False

    def itemChange(self, change, value):
        if change == QGraphicsItem.ItemPositionChange and self.scene():
            newPos = value.toPoint()
            sceneRect = self.scene().sceneRect()
            viewRect = self.scene().views()[0].viewport().rect()
            newPos.setX(round(newPos.x() / self._gridTuple[0]) * self._gridTuple[0])
            newPos.setY(round(newPos.y() / self._gridTuple[1]) * self._gridTuple[1])

            if not sceneRect.contains(newPos):
                # Keep the item inside the scene rect.
                if newPos.x() > sceneRect.right():
                    sceneRect.setRight(newPos.x())
                    viewRect.setRight(newPos.x())
                elif newPos.x() < sceneRect.left():
                    sceneRect.setLeft(newPos.x())
                    viewRect.setLeft(newPos.x())
                if newPos.y() > sceneRect.bottom():
                    sceneRect.setBottom(newPos.y())
                    viewRect.setBottom(newPos.y())
                elif newPos.y() < sceneRect.top():
                    sceneRect.setTop(newPos.y())
                    viewRect.setTop(newPos.y())
            return newPos
        return super().itemChange(change, value)

    @property
    def angle(self):
        return self._angle

    @angle.setter
    def angle(self, value):
        self._angle = value
        self.prepareGeometryChange()
        self.setRotation(value)  # self.update(self.boundingRect())

    @property
    def pen(self):
        return self._pen

    @pen.setter
    def pen(self, value):
        self._pen = value  # self.update(self.boundingRect())

    @property
    def gridTuple(self):
        return self._gridTuple

    @gridTuple.setter
    def gridTuple(self, value: int):
        self._gridTuple = value

    @property
    def stretch(self):
        return self._stretch

    @stretch.setter
    def stretch(self, value: bool):
        self._stretch = value  # self.update(self.boundingRect())

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mousePressEvent(event)
        if self.scene().changeOrigin:
            self.setFlag(QGraphicsItem.ItemIsMovable, False)
            self.setFlag(QGraphicsItem.ItemIsSelectable, False)
        else:
            self.setFlag(QGraphicsItem.ItemIsMovable, True)
            self.setFlag(QGraphicsItem.ItemIsSelectable, True)

    def sceneEvent(self, event):
        """
        Do not propagate event if shape needs to keep still.
        """

        if self.scene() and (self.scene().changeOrigin or self.scene().drawMode):
            return False
        else:
            super().sceneEvent(event)
            return True

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseReleaseEvent(event)  # self.setSelected(False)

    def hoverEnterEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        self.setCursor(Qt.ArrowCursor)
        self.setOpacity(0.75)
        self.setFocus()
        super().hoverEnterEvent(event)

    def hoverLeaveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().hoverLeaveEvent(event)
        self.setCursor(Qt.CrossCursor)
        self.setOpacity(1)
        self.clearFocus()

    def contextMenuEvent(self, event):
        self.scene().itemContextMenu.exec_(event.screenPos())

    def snapToBase(self, number, base):
        """
        Restrict a number to the multiples of base
        """
        return base * int(round(number / base))

    def snapToGrid(self, point: QPoint, gridTuple: tuple[int, int]):
        """
        snap point to grid. Divides and multiplies by grid size.
        """
        return QPoint(gridTuple[0] * int(round(point.x() / gridTuple[0])),
                      gridTuple[1] * int(round(point.y() / gridTuple[1])), )


class rectangle(shape):
    """
    rect: QRect defined by top left corner and bottom right corner. QRect(Point1,Point2)
    """

    sides = ["Left", "Right", "Top", "Bottom"]

    def __init__(self, start: QPoint, end: QPoint, pen: QPen, grid: tuple, ):
        super().__init__(pen, grid)
        self._rect = QRectF(start, end).normalized()
        self._start = self._rect.topLeft()
        self._end = self._rect.bottomRight()
        self._stretchSide = None

    def boundingRect(self):
        return self._rect.normalized().adjusted(-2, -2, 2, 2)

    def paint(self, painter, option, widget):
        if self.isSelected():
            painter.setPen(QPen(Qt.yellow, 2, Qt.DashLine))
            painter.drawRect(self._rect)
            if self.stretch:
                painter.setPen(QPen(Qt.red, 2, Qt.SolidLine))
                if self._stretchSide == rectangle.sides[0]:
                    painter.drawLine(self.rect.topLeft(), self.rect.bottomLeft())
                elif self._stretchSide == rectangle.sides[1]:
                    painter.drawLine(self.rect.topRight(), self.rect.bottomRight())
                elif self._stretchSide == rectangle.sides[2]:
                    painter.drawLine(self.rect.topLeft(), self.rect.topRight())
                elif self._stretchSide == rectangle.sides[3]:
                    painter.drawLine(self.rect.bottomLeft(), self.rect.bottomRight())
        else:
            painter.setPen(self._pen)
            painter.drawRect(self._rect)

    @property
    def rect(self):
        return self._rect

    @rect.setter
    def rect(self, rect: QRect):
        self.prepareGeometryChange()
        self._rect = rect

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, start: QPoint):
        self.prepareGeometryChange()
        self._rect = QRectF(start, self.end).normalized()
        self._start = start

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, end: QPoint):
        self.prepareGeometryChange()
        self._rect = QRectF(self.start, end).normalized()
        self._end = end

    @property
    def centre(self):
        return QPoint(int(self._rect.x() + self._rect.width() / 2),
                      int(self._rect.y() + self._rect.height() / 2), )

    @property
    def height(self):
        return self.rect.height()

    @height.setter
    def height(self, height: int):
        self.prepareGeometryChange()
        self._rect.setHeight(height)

    @property
    def width(self):
        return self.rect.width()

    @width.setter
    def width(self, width):
        self.prepareGeometryChange()
        self.rect.setWidth(width)

    @property
    def objName(self):
        return "RECTANGLE"

    @property
    def left(self):
        return self.rect.left()

    @left.setter
    def left(self, left: int):
        self.rect.setLeft(left)

    @property
    def right(self):
        return self.rect.right()

    @right.setter
    def right(self, right: int):
        self.prepareGeometryChange()
        self.rect.setRight(right)

    @property
    def top(self):
        return self.rect.top()

    @top.setter
    def top(self, top: int):
        self.prepareGeometryChange()
        self.rect.setTop(top)

    @property
    def bottom(self):
        return self.rect.bottom()

    @bottom.setter
    def bottom(self, bottom: int):
        self.prepareGeometryChange()
        self.rect.setBottom(bottom)

    @property
    def origin(self):
        return self.rect.bottomLeft()

    @property
    def stretchSide(self):
        return self._stretchSide

    @stretchSide.setter
    def stretchSide(self, value: str):
        self.prepareGeometryChange()
        self._stretchSide = value

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mousePressEvent(event)
        eventPos = event.pos().toPoint()
        if self._stretch:
            self.setFlag(QGraphicsItem.ItemIsMovable, False)
            if eventPos.x() == self.snapToBase(self._rect.left(), self.gridTuple[0]):
                if self._rect.top() <= eventPos.y() <= self._rect.bottom():
                    self.setCursor(Qt.SizeHorCursor)
                    self._stretchSide = rectangle.sides[0]
            elif eventPos.x() == self.snapToBase(self._rect.right(), self.gridTuple[0]):
                if self._rect.top() <= eventPos.y() <= self._rect.bottom():
                    self.setCursor(Qt.SizeHorCursor)
                    self._stretchSide = rectangle.sides[1]
            elif eventPos.y() == self.snapToBase(self._rect.top(), self.gridTuple[1]):
                if self._rect.left() <= eventPos.x() <= self._rect.right():
                    self.setCursor(Qt.SizeVerCursor)
                    self._stretchSide = rectangle.sides[2]
            elif eventPos.y() == self.snapToBase(self._rect.bottom(), self.gridTuple[1]):
                if self._rect.left() <= eventPos.x() <= self._rect.right():
                    self.setCursor(Qt.SizeVerCursor)
                    self._stretchSide = rectangle.sides[3]

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        eventPos = event.pos().toPoint()
        if self.stretch:
            self.prepareGeometryChange()
            if self.stretchSide == rectangle.sides[0]:
                self.setCursor(Qt.SizeHorCursor)
                self.rect.setLeft(eventPos.x())
            elif self.stretchSide == rectangle.sides[1]:
                self.setCursor(Qt.SizeHorCursor)
                self.rect.setRight(eventPos.x() - int(self.pen.width() / 2))
            elif self.stretchSide == rectangle.sides[2]:
                self.setCursor(Qt.SizeVerCursor)
                self.rect.setTop(eventPos.y())
            elif self.stretchSide == rectangle.sides[3]:
                self.setCursor(Qt.SizeVerCursor)
                self.rect.setBottom(eventPos.y() - int(self.pen.width() / 2))
            self.update()
        else:
            super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        self.setFlag(QGraphicsItem.ItemIsMovable, True)
        super().mouseReleaseEvent(event)
        # self.start = self.rect.topLeft()
        # self.end = self.rect.bottomRight()
        if self.stretch:
            self._stretch = False
            self._stretchSide = None
            self.setCursor(Qt.ArrowCursor)


class circle(shape):
    def __init__(self, centre: QPoint, end: QPoint, pen: QPen, gridTuple: tuple):
        super().__init__(pen, gridTuple)
        xlen = abs(end.x() - centre.x())
        ylen = abs(end.y() - centre.y())
        self._radius = self.snapToBase(int(math.sqrt(xlen ** 2 + ylen ** 2)),
                                       self.gridTuple[0])
        self._centre = centre
        self._topLeft = self._centre - QPoint(self._radius, self._radius)
        self._rightBottom = self._centre + QPoint(self._radius, self._radius)
        self._end = self._centre + QPoint(self._radius, 0)  # along x-axis
        self._stretch = False
        self._startStretch = False

    def paint(self, painter, option, widget) -> None:
        if self.isSelected():
            painter.setPen(QPen(Qt.yellow, 2, Qt.DashLine))
            painter.drawEllipse(self._centre, 1, 1)
            if self._stretch:
                painter.setPen(QPen(Qt.red, 2, Qt.SolidLine))
        else:
            painter.setPen(self._pen)
        painter.drawEllipse(self._centre, self._radius, self._radius)

    @property
    def centre(self):
        return self._centre

    @centre.setter
    def centre(self, centre: QPoint):
        self.prepareGeometryChange()
        self._centre = self.snapToGrid(centre, self._gridTuple)
        # self.topLeft = self._centre - QPoint(self._radius, self._radius)
        # self.rightBottom = self._centre + QPoint(self._radius, self._radius)
        self._end = self._centre + QPoint(self._radius, 0)

    @property
    def radius(self):
        return self._radius

    @radius.setter
    def radius(self, radius: int):
        self.prepareGeometryChange()
        self._radius = self.snapToBase(radius, self._gridTuple[0])
        self._end = self._centre + QPoint(self._radius, 0)
        self._topLeft = self._centre - QPoint(self._radius, self._radius)
        self._rightBottom = self._centre + QPoint(self._radius, self._radius)

    @property
    def centre(self):
        return self._centre

    @centre.setter
    def centre(self, value: QPoint):
        self.prepareGeometryChange()
        if isinstance(value, QPoint):
            self._centre = value

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, value: QPoint):
        if isinstance(value, QPoint):
            self.prepareGeometryChange()
            self._end = value

    @property
    def rightBottom(self):
        return self._rightBottom

    @rightBottom.setter
    def rightBottom(self, value: QPoint):
        if isinstance(value, QPoint):
            self._rightBottom = value

    @property
    def topLeft(self):
        return self._topLeft

    @topLeft.setter
    def topLeft(self, value: QPoint):
        if isinstance(value, QPoint):
            self._topLeft = value

    @property
    def objName(self):
        return "CIRCLE"

    def boundingRect(self):
        return (
            QRectF(self._topLeft, self._rightBottom).normalized().adjusted(-2, -2, 2, 2))

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mousePressEvent(event)
        if self.isSelected() and self._stretch:
            self.setFlag(QGraphicsItem.ItemIsMovable, False)
            # eventPos = self.snap2grid(event.pos(), self._gridTuple)
            eventPos = event.pos().toPoint()
            distance = self.snapToBase(math.sqrt((eventPos.x() - self._centre.x()) ** 2 + (
                    eventPos.y() - self._centre.y()) ** 2), self._gridTuple[0], )
            if distance == self._radius:
                self._startStretch = True
                self.setCursor(Qt.DragMoveCursor)

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseMoveEvent(event)
        if self._startStretch:
            eventPos = event.pos().toPoint()
            distance = self.snapToBase(math.sqrt((eventPos.x() - self._centre.x()) ** 2 + (
                    eventPos.y() - self._centre.y()) ** 2), self._gridTuple[0], )
            self.prepareGeometryChange()
            self._radius = distance

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseReleaseEvent(event)
        if self._startStretch:
            self._startStretch = False
            self.setFlag(QGraphicsItem.ItemIsMovable, True)
            self.setFlag(QGraphicsItem.ItemIsSelectable, True)
            self._topLeft = self._centre - QPoint(self._radius, self._radius)
            self._rightBottom = self._centre + QPoint(self._radius, self._radius)
            self._end = self._centre + QPoint(self._radius, 0)
            self.setCursor(Qt.ArrowCursor)


class arc(shape):
    """
    Class to draw arc shapes. Can have four directions.
    """

    arcTypes = ["Up", "Right", "Down", "Left"]
    sides = ["Left", "Right", "Top", "Bottom"]

    def __init__(self, start: QPoint, end: QPoint, pen: QPen, gridTuple: tuple):
        super().__init__(pen, gridTuple)
        self._start = start
        self._end = end
        self._rect = QRectF(self._start, self._end).normalized()
        self._arcLine = QLineF(self._start, self._end)
        self._arcAngle = 0
        self._width = self._rect.width()
        self._height = self._rect.height()
        self._adjustment = int(self.pen.width() / 2)
        self._stretchSide = None
        self.findAngle()

    def findAngle(self):
        self._arcAngle = self._arcLine.angle()
        if 90 >= self._arcAngle >= 0:
            self._arcType = arc.arcTypes[0]
        elif 180 >= self._arcAngle > 90:
            self._arcType = arc.arcTypes[1]
        elif 270 >= self._arcAngle > 180:
            self._arcType = arc.arcTypes[2]
        elif 360 > self._arcAngle > 270:
            self._arcType = arc.arcTypes[3]

    def paint(self, painter, option, widget) -> None:
        if self.isSelected():
            painter.setPen(QPen(Qt.yellow, 2, Qt.DashLine))
            painter.drawRect(self.rect)
            self.arcDraw(painter)
            if self._stretch:
                painter.setPen(QPen(Qt.red, 2, Qt.SolidLine))
                self.arcDraw(painter)
                if self._stretchSide == arc.sides[0]:
                    painter.drawLine(self._rect.topLeft(), self._rect.bottomLeft())
                elif self._stretchSide == arc.sides[1]:
                    painter.drawLine(self._rect.topRight(), self._rect.bottomRight())
                elif self._stretchSide == arc.sides[2]:
                    painter.drawLine(self._rect.topLeft(), self._rect.topRight())
                elif self._stretchSide == arc.sides[3]:
                    painter.drawLine(self._rect.bottomLeft(), self._rect.bottomRight())
        else:
            painter.setPen(self._pen)
            self.arcDraw(painter)

    def arcDraw(self, painter):
        if self._arcType == arc.arcTypes[0]:
            painter.drawArc(self._rect, 0, 180 * 16)
        elif self._arcType == arc.arcTypes[1]:
            painter.drawArc(self._rect, 90 * 16, 180 * 16)
        elif self._arcType == arc.arcTypes[2]:
            painter.drawArc(self._rect, 180 * 16, 180 * 16)
        elif self._arcType == arc.arcTypes[3]:
            painter.drawArc(self._rect, 270 * 16, 180 * 16)

    def boundingRect(self):
        return self._rect.adjusted(-2, -2, 2, 2)

    @property
    def objName(self):
        return "ARC"

    @property
    def rect(self):
        return self._rect

    @rect.setter
    def rect(self, arc_rect: QRect):
        assert isinstance(arc_rect, QRect)
        self._rect = arc_rect.normalized()

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, point: QPoint):
        assert isinstance(point, QPoint)
        self.prepareGeometryChange()
        self._start = self.snapToGrid(point, self.gridTuple)

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, point: QPoint):
        assert isinstance(point, QPoint)
        self.prepareGeometryChange()
        self._end = point
        self._arcLine = QLineF(self._start, self._end)
        self._arcAngle = self._arcLine.angle()
        self.findAngle()
        self._rect = QRectF(self._start, self._end).normalized()

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, width):
        self._width = width
        self.prepareGeometryChange()
        self.rect.setWidth(self._width)

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, height):
        self._height = height
        self.prepareGeometryChange()
        self.rect.setHeight(self._height)

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mousePressEvent(event)
        eventPos = event.pos().toPoint()
        if self._stretch:
            self.setFlag(QGraphicsItem.ItemIsMovable, False)
            if eventPos.x() == self.snapToBase(self._rect.left(), self.gridTuple[0]):
                if self._rect.top() <= eventPos.y() <= self._rect.bottom():
                    self.setCursor(Qt.SizeHorCursor)
                    self._stretchSide = arc.sides[0]
            elif eventPos.x() == self.snapToBase(self._rect.right(), self.gridTuple[0]):
                if self._rect.top() <= eventPos.y() <= self._rect.bottom():
                    self.setCursor(Qt.SizeHorCursor)
                    self._stretchSide = arc.sides[1]
            elif eventPos.y() == self.snapToBase(self._rect.top(), self.gridTuple[1]):
                if self._rect.left() <= eventPos.x() <= self._rect.right():
                    self.setCursor(Qt.SizeVerCursor)
                    self._stretchSide = arc.sides[2]
            elif eventPos.y() == self.snapToBase(self._rect.bottom(), self.gridTuple[1]):
                if self._rect.left() <= eventPos.x() <= self._rect.right():
                    self.setCursor(Qt.SizeVerCursor)
                    self._stretchSide = arc.sides[3]

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseMoveEvent(event)

        eventPos = event.pos().toPoint()
        if self._stretch:
            self.prepareGeometryChange()
            if self._stretchSide == arc.sides[0]:
                self.setCursor(Qt.SizeHorCursor)
                self._rect.setLeft(eventPos.x())
            elif self._stretchSide == arc.sides[1]:
                self.setCursor(Qt.SizeHorCursor)
                self._rect.setRight(eventPos.x() - self._adjustment)
            elif self._stretchSide == arc.sides[2]:
                self.setCursor(Qt.SizeVerCursor)
                self._rect.setTop(eventPos.y())
            elif self._stretchSide == arc.sides[3]:
                self.setCursor(Qt.SizeVerCursor)
                self._rect.setBottom(eventPos.y() - self._adjustment)
            self.update()

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        self.setFlag(QGraphicsItem.ItemIsMovable, True)
        super().mouseReleaseEvent(event)
        if self.stretch:
            self._stretch = False
            self._stretchSide = None
            self.setCursor(Qt.ArrowCursor)

            if self._arcType == arc.arcTypes[0]:
                self._start = self.snapToGrid(self._rect.bottomLeft(), self.gridTuple)
                self._end = self.snapToGrid(self._rect.topRight(), self.gridTuple)
            elif self._arcType == arc.arcTypes[1]:
                self._start = self.snapToGrid(self._rect.topLeft(), self.gridTuple)
                self._end = self.snapToGrid(self._rect.bottomRight(), self.gridTuple)
            elif self._arcType == arc.arcTypes[2]:
                self._start = self.snapToGrid(self._rect.topRight(), self.gridTuple)
                self._end = self.snapToGrid(self._rect.bottomLeft(), self.gridTuple)
            elif self._arcType == arc.arcTypes[3]:
                self._start = self.snapToGrid(self._rect.bottomRight(), self.gridTuple)
                self._end = self.snapToGrid(self._rect.topLeft(), self.gridTuple)
            self._rect = QRectF(self._start, self._end).normalized()


class line(shape):
    """
    line class definition for symbol drawing.
    """

    stretchSides = ["start", "end"]

    def __init__(self, start: QPoint, end: QPoint, pen: QPen, grid: tuple, ):
        super().__init__(pen, grid)
        self._end = end
        self._start = start
        self._pen = pen
        self._stretch = False
        self._stretchSide = None
        self._line = QLine(self._start, self._end)
        self._rect = QRect(self._start, self._end).normalized()
        self._horizontal = True  # True if line is horizontal, False if vertical

    def boundingRect(self):
        return self._rect.adjusted(-2, -2, 2, 2)

    def shape(self):
        path = QPainterPath()
        path.addRect(self._rect.adjusted(-2, -2, 2, 2))
        return path

    def shape(self):
        path = QPainterPath()
        path.addRect(self._rect.adjusted(-2, -2, 2, 2))
        return path

    def paint(self, painter, option, widget):
        if self.isSelected():
            painter.setPen(QPen(Qt.yellow, 2, Qt.DashLine))
            if self._stretch:
                painter.setPen(QPen(Qt.red, 2, Qt.SolidLine))
                if self._stretchSide == line.stretchSides[0]:
                    painter.drawEllipse(self._start, self.gridTuple[0], self.gridTuple[1])
                elif self._stretchSide == line.stretchSides[1]:
                    painter.drawEllipse(self._end, self.gridTuple[0], self.gridTuple[1])
        else:
            painter.setPen(self._pen)
        painter.drawLine(self._line)

    def objName(self):
        return "LINE"

    @property
    def rect(self):
        return self._rect

    @rect.setter
    def rect(self, rect: QRect):
        self._rect = rect

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, start: QPoint):
        self.prepareGeometryChange()
        self._start = start
        self._line = QLine(self._start, self._end)
        self._rect = QRect(self._start, self._end).normalized()

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, end: QPoint):
        self.prepareGeometryChange()
        self._end = end
        self._line = QLine(self._start, self._end)
        self._rect = QRect(self._start, self._end).normalized()

    @property
    def pen(self):
        return self._pen

    @pen.setter
    def pen(self, pen: QPen):
        self._pen = pen

    @property
    def width(self):
        return self._pen.width()

    @width.setter
    def width(self, width: int):
        self._pen.setWidth(width)

    def bBox(self) -> QRect:
        return self.boundingRect()

    def Move(self, offset: QPoint):
        self.start += offset
        self._end += offset

    @property
    def length(self):
        return math.sqrt(
            (self.start.x() - self._end.x()) ** 2 + (self.start.y() - self._end.y()) ** 2)

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mousePressEvent(event)
        if self.isSelected() and self._stretch:
            self.setFlag(QGraphicsItem.ItemIsMovable, False)
            eventPos = event.pos().toPoint()
            if eventPos == self.start:
                self._stretchSide = line.stretchSides[0]
            elif eventPos == self._end:
                self._stretchSide = line.stretchSides[1]

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        eventPos = event.pos().toPoint()
        if self._stretchSide == line.stretchSides[0]:
            self.prepareGeometryChange()
            self.start = eventPos
            self._line = QLine(self.start, self._end)
            self._rect = QRect(self.start, self._end).normalized()
        elif self._stretchSide == line.stretchSides[1]:
            self.prepareGeometryChange()
            self._end = eventPos
            self._line = QLine(self.start, self._end)
            self._rect = QRect(self.start, self._end).normalized()

        super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseReleaseEvent(event)
        self._stretch = False
        self.setFlag(QGraphicsItem.ItemIsMovable, True)
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)
        self._stretchSide = ""


class pin(shape):
    """
    symbol pin class definition for symbol drawing.
    """

    pinDirs = ["Input", "Output", "Inout"]
    pinTypes = ["Signal", "Ground", "Power", "Clock", "Digital", "Analog"]

    def __init__(self, start: QPoint, pen: QPen, pinName: str = "",
                 pinDir: str = pinDirs[0], pinType: str = pinTypes[0],
                 grid: tuple = (10, 10), ):
        super().__init__(pen, grid)

        self._start = start  # centre of pin
        self._pinName = pinName
        self._pinDir = pinDir
        self._pinType = pinType
        self._connected = False  # True if the pin is connected to a net.
        self._rect = QRect(self._start.x() - 5, self._start.y() - 5, 10, 10)

    def __repr__(self):
        return f"pin: {self._pinName} {self.mapToScene(self._start)}"

    def boundingRect(self):
        return self._rect  #

    def paint(self, painter, option, widget):
        painter.setPen(self._pen)
        painter.setBrush(self._pen.color())
        painter.drawRect(self._rect)
        painter.setFont(QFont("Arial", 12))
        painter.drawText(QPoint(self._start.x() - 5, self._start.y() - 10), self._pinName)
        if self.isSelected():
            painter.setPen(QPen(Qt.yellow, 2, Qt.DashLine))
            painter.setBrush(Qt.yellow)
            painter.drawRect(self._rect)

    def objName(self):
        return "PIN"

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, start):
        self._start = start

    @property
    def pinName(self):
        return self._pinName

    @pinName.setter
    def pinName(self, pinName):
        if pinName != "":
            self._pinName = pinName

    @property
    def pinDir(self):
        return self._pinDir

    @pinDir.setter
    def pinDir(self, direction: str):
        if direction in self.pinDirections:
            self._pinDir = direction

    @property
    def pinType(self):
        return self._pinType

    @pinType.setter
    def pinType(self, pintype: str):
        if pintype in self.pinTypes:
            self._pinType = pintype

    @property
    def connected(self):
        return self._connected

    @connected.setter
    def connected(self, value: bool):
        if isinstance(value, bool):
            self._connected = value

    def toSchematicPin(self, start: QPoint, pen: QPen, gridTuple):
        return schematicPin(start, pen, self.pinName, self.pinDir, self.pinType, gridTuple)


class text(shape):
    """
    This class is for text annotations on symbol or schematics.
    """

    textAlignments = ["Left", "Center", "Right"]
    textOrients = ["R0", "R90", "R180", "R270"]

    def __init__(self, start: QPoint, pen: QPen, textContent: str = "",
                 grid: tuple = (10, 10), fontFamily="Helvetica", fontStyle="Regular",
                 textHeight: str = "12", textAlign: str = "Left", textOrient: str = "R0", ):
        super().__init__(pen, grid)

        self._start = start
        self._pen = pen
        self._textContent = textContent
        self._textHeight = textHeight
        self._textAlign = textAlign
        self._textOrient = textOrient
        self._textFont = QFont(fontFamily)
        self._textFont.setStyleName(fontStyle)
        self._textFont.setPointSize(int(float(self._textHeight)))
        self._textFont.setKerning(True)
        self.setOpacity(1)
        self._fm = QFontMetrics(self._textFont)
        self._textOptions = QTextOption()
        if self._textAlign == text.textAlignments[0]:
            self._textOptions.setAlignment(Qt.AlignmentFlag.AlignLeft)
        elif self._textAlign == text.textAlignments[1]:
            self._textOptions.setAlignment(Qt.AlignmentFlag.AlignCenter)
        elif self._textAlign == text.textAlignments[2]:
            self._textOptions.setAlignment(Qt.AlignmentFlag.AlignRight)

    def boundingRect(self):
        if self._textAlign == text.textAlignments[0]:
            self._rect = self._fm.boundingRect(QRect(0, 0, 400, 400),
                                               Qt.AlignmentFlag.AlignLeft,
                                               self._textContent)
        elif self._textAlign == text.textAlignments[1]:
            self._rect = self._fm.boundingRect(QRect(0, 0, 400, 400),
                                               Qt.AlignmentFlag.AlignCenter,
                                               self._textContent)
        elif self._textAlign == text.textAlignments[2]:
            self._rect = self._fm.boundingRect(QRect(0, 0, 400, 400),
                                               Qt.AlignmentFlag.AlignRight,
                                               self._textContent)
        return QRect(self._start.x(), self._start.y() - self._rect.height(),
                     self._rect.width(), self._rect.height(), )

    def paint(self, painter, option, widget):
        painter.setFont(self._textFont)
        if self.isSelected():
            painter.setPen(QPen(Qt.yellow, 2, Qt.DashLine))
            painter.drawRect(self.boundingRect())
        else:
            painter.setPen(self._pen)
        painter.drawText(self.boundingRect(), self._textContent, o=self._textOptions)

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, value: QPoint):
        self._start = value

    @property
    def textContent(self):
        return self._textContent

    @textContent.setter
    def textContent(self, inputText: str):
        if isinstance(inputText, str):
            self._textContent = inputText
        else:
            self.scene().logger.error(f"Not a string: {inputText}")

    @property
    def fontFamily(self) -> str:
        return self._textFont.family()

    @fontFamily.setter
    def fontFamily(self, familyName):
        fontFamilies = QFontDatabase.families(QFontDatabase.Latin)
        fixedFamilies = [family for family in fontFamilies if
                         QFontDatabase.isFixedPitch(family)]
        if familyName in fixedFamilies:
            self._textFont.setFamily(familyName)
        else:
            self.scene().logger.error(f"Not a valid font name: {familyName}")

    @property
    def fontStyle(self) -> str:
        return self._textFont.styleName()

    @fontStyle.setter
    def fontStyle(self, value: str):
        if value in QFontDatabase.styles(self._textFont.family()):
            self._textFont.setStyleName(value)
        else:
            self.scene().logger.error(f"Not a valid font style: {value}")

    @property
    def textHeight(self) -> int:
        return self._textHeight

    @textHeight.setter
    def textHeight(self, value: int):
        fontSizes = [str(size) for size in QFontDatabase.pointSizes(self._textFont.family(),
                                                                    self._textFont.styleName())]
        if value in fontSizes:
            self._textHeight = value
        else:
            self.scene().logger.error(f"Not a valid font height: {value}")
            self.scene().logger.warning(f"Valid font heights are: {fontSizes}")

    @property
    def textFont(self) -> QFont:
        return self._textFont

    @textFont.setter
    def textFont(self, value: QFont):
        assert isinstance(value, QFont)
        self._textFont = value

    @property
    def textAlignment(self):
        return self._textAlign

    @textAlignment.setter
    def textAlignment(self, value):
        if value in text.textAlignments:
            self._textAlign = value
        else:
            self.scene().logger.error(f"Not a valid text alignment value: {value}")

    @property
    def textOrient(self):
        return self._textOrient

    @textOrient.setter
    def textOrient(self, value):
        if value in text.textOrients:
            self._textOrient = value
        else:
            self.scene().logger.error(f"Not a valid text orientation: {value}")


class label(shape):
    """
    label: text class definition for symbol drawing.
    labelText is what is shown on the symbol in a schematic
    """

    labelAlignments = ["Left", "Center", "Right"]
    labelOrients = ["R0", "R90", "R180", "R270", "MX", "MX90", "MY", "MY90"]
    labelUses = ["Normal", "Instance", "Pin", "Device", "Annotation"]
    labelTypes = ["Normal", "NLPLabel", "PyLabel"]
    predefinedLabels = ["[@libName]", "[@cellName]", "[@viewName]", "[@instName]",
                        "[@modelName]", "[@elementNum]"]

    def __init__(self, start: QPoint, pen: QPen, labelDefinition: str = "",
                 grid: tuple = (10, 10), labelType: str = "Normal", labelHeight: str = "12",
                 labelAlign: str = "Left", labelOrient: str = "R0",
                 labelUse: str = "Normal", ):
        super().__init__(pen, grid)
        self._start = start  # top left corner
        self._pen = pen
        self._labelDefinition = labelDefinition  # label definition is what is
        # entered in the symbol editor
        self._labelName = None  # label Name
        self._labelValue = None  # label value
        self._labelText = None  # Displayed label
        self._labelHeight = labelHeight
        self._labelAlign = labelAlign
        self._labelOrient = labelOrient
        self._labelUse = labelUse
        self._labelType = labelType
        self._labelFont = QFont("Arial")
        self._labelFont.setPointSize(int(float(self._labelHeight)))
        self._labelFont.setKerning(False)
        self._labelVisible: bool = False
        self._labelValueSet: bool = False
        # labels are visible by default
        self.setOpacity(1)
        self._fm = QFontMetrics(self._labelFont)
        self._rect = self._fm.boundingRect(self._labelDefinition)

    def __repr__(self):
        return f"label: {self._labelText} for {self._labelDefinition} at {self._start}, " \
               f"value =  {self._labelValue}"

    def boundingRect(self):
        return QRect(self._start.x(), self._start.y(), self._rect.width(),
                     self._rect.height()).normalized().adjusted(0, 0, 0, 5)  #

    def shape(self) -> QPainterPath:
        path = QPainterPath()
        path.addRect(self.boundingRect())
        return path

    def paint(self, painter, option, widget):
        # self._rect = self.fm.boundingRect(self.labelName)
        self._labelFont.setPointSize(int(self._labelHeight))
        painter.setFont(self._labelFont)
        if self.isSelected():
            painter.setPen(QPen(Qt.yellow, 2, Qt.DashLine))
            painter.drawRect(self.boundingRect())
        else:
            painter.setPen(self._pen)
        if self._labelText:
            painter.drawText(QPoint(self._start.x(), self._start.y() + self._rect.height()),
                self._labelText, )
        else:
            painter.drawText(QPoint(self._start.x(), self._start.y() + self._rect.height()),
                self._labelDefinition, )
        self._fm = QFontMetrics(self._labelFont)
        self._rect = self._fm.boundingRect(self._labelDefinition)

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, value: QPoint):
        self._start = value

    @property
    def left(self):
        return self._start.x()

    @property
    def right(self):
        return self._start.x() + self.boundingRect().width()

    @property
    def top(self):
        return self._start.y()

    @property
    def bottom(self):
        return self._start.y() + self.boundingRect().height()

    @property
    def width(self):
        return self.boundingRect().width()

    @property
    def height(self):
        return self.boundingRect().height()

    @property
    def labelName(self):
        return self._labelName

    @labelName.setter
    def labelName(self, labelName: str):
        self._labelName = labelName

    @property
    def labelDefinition(self):
        return self._labelDefinition

    @labelDefinition.setter
    def labelDefinition(self, labelDefinition: str):
        if isinstance(labelDefinition, str):
            self._labelDefinition = labelDefinition.strip()

    @property
    def labelValue(self):
        return self._labelValue

    @labelValue.setter
    def labelValue(self, labelValue):
        self._labelValue = labelValue
        self._labelValueSet = True  # self.labelDefs()

    @property
    def labelValueSet(self) -> bool:
        return self._labelValueSet

    @labelValueSet.setter
    def labelValueSet(self, value: bool):
        if isinstance(value, bool):
            self._labelValueSet = value

    @property
    def labelHeight(self):
        return self._labelHeight

    @labelHeight.setter
    def labelHeight(self, value: int):
        self._labelHeight = value

    @property
    def labelText(self):
        return self._labelText

    @labelText.setter
    def labelText(self, labelText):
        self._labelText = labelText
        self._rect = self._fm.boundingRect(self._labelText).normalized().adjusted(0, 0, 0,
                                                                                  5)

    def objName(self):
        return "LABEL"

    @property
    def labelType(self):
        return self._labelType

    @labelType.setter
    def labelType(self, labelType):
        if labelType in self.labelTypes:
            self._labelType = labelType
        elif self.scene():
            self.scene().logger.error("Invalid label type")

    @property
    def labelAlign(self):
        return self._labelAlign

    @labelAlign.setter
    def labelAlign(self, labelAlignment):
        if labelAlignment in self.labelAlignments:
            self._labelAlign = labelAlignment
        elif self.scene():
            self.scene().logger.error("Invalid label alignment")

    @property
    def labelOrient(self):
        return self._labelOrient

    @labelOrient.setter
    def labelOrient(self, labelOrient):
        if labelOrient in self.labelOrients:
            self._labelOrient = labelOrient
        else:
            print("Invalid label orientation")

    @property
    def labelUse(self):
        return self._labelUse

    @labelUse.setter
    def labelUse(self, labelUse):
        if labelUse in self.labelUses:
            self._labelUse = labelUse
        elif self.scene():
            self.scene().logger.error("Invalid label use")

    @property
    def labelFont(self):
        return self._labelFont

    @labelFont.setter
    def labelFont(self, labelFont: QFont):
        self._labelFont = labelFont

    @property
    def labelVisible(self) -> bool:
        return self._labelVisible

    @labelVisible.setter
    def labelVisible(self, value: bool):
        assert isinstance(value, bool)
        if value:
            self.setOpacity(1)
            self._labelVisible = True
        else:
            self.setOpacity(0.001)
            self._labelVisible = False

    def moveBy(self, delta: QPoint):
        self._start += delta

    def labelDefs(self):
        """
        This method will create label name, value andtext from label
        definition. It should be run label is defined or redefined.
        """
        self.prepareGeometryChange()
        if self._labelType == label.labelTypes[0]:
            self._labelName = self._labelDefinition
            self._labelText = self._labelDefinition
            self._labelValue = None
            self._labelValueSet = True
        elif self.labelType == label.labelTypes[1]:
            try:
                if self._labelDefinition in label.predefinedLabels:
                    self._labelValueSet = True
                    match self.labelDefinition:
                        case "[@cellName]":
                            self._labelName = "cellName"
                            self._labelValue = self.parentItem().cellName
                            self._labelText = self._labelValue
                        case "[@instName]":
                            self._labelName = "instName"
                            self._labelValue = f"I{self.parentItem().counter}"
                            self._labelText = self._labelValue
                        case "[@libName]":
                            self._labelName = "libName"
                            self._labelValue = self.parentItem().libraryName
                            self._labelText = self._labelValue
                        case "[@viewName]":
                            self._viewName = "viewName"
                            self._labelValue = self.parentItem().viewName
                            self._labelText = self._labelValue
                        case "[@modelName]":
                            self._labelName = "modelName"
                            self._labelValue = self.parentItem().attr.get("modelName", "")
                            self._labelText = self._labelValue
                        case "[@elementNum]":
                            self._labelName = "elementNum"
                            self._labelValue = f"{self.parentItem().counter}"
                            self._labelText = self._labelValue
                else:
                    labelFields = (
                        self._labelDefinition.lstrip("[@").rstrip("]").rstrip(":").split(
                            ":"))
                    self._labelName = labelFields[0].strip()
                    match len(labelFields):
                        case 1:
                            if not self._labelValueSet:
                                self._labelValue = "?"
                            self._labelText = self._labelValue
                        case 2:
                            if self._labelValueSet:
                                self._labelText = (
                                    labelFields[1].strip().replace("%", self._labelValue))
                            else:
                                self._labelValue = "?"
                        case 3:
                            tempLabelValue = (
                                labelFields[2].strip().split("=")[-1].split()[-1])
                            if self.labelValueSet:
                                self._labelText = labelFields[2].replace(tempLabelValue,
                                    self._labelValue)
                            else:
                                self._labelText = labelFields[2]
                                self._labelValue = tempLabelValue

            except Exception as e:
                if self.scene():
                    self.scene().logger.error(e)
        elif self._labelType == label.labelTypes[2]:  # pyLabel
            try:
                labelFields = self._labelDefinition.strip().split("=")
                self._labelName = labelFields[0].strip()
                labelFunction = labelFields[1].strip()
                # pass the PDK callback class named with "cellName" the labels
                # dictionary of instance.w
                expression = (
                    f"cb.{self.parentItem().cellName}(self.parentItem().labels).{labelFunction}")
                self._labelValue = Quantity(eval(expression)).render(prec=3)
                self._labelText = f"{self._labelName}={self._labelValue}"
            except Exception as e:
                if self.scene():
                    self.scene().logger.error(e)


class symbolShape(shape):
    def __init__(self, pen: QPen, gridTuple: tuple, shapes: list, attr: dict):
        super().__init__(pen, gridTuple)
        assert shapes is not None  # must not be an empty list
        self.shapes = shapes  # list of shapes in the symbol
        self.attr = attr  # parameters common to all instances of symbol
        self._counter = 0  # item's number on schematic
        self._libraryName = ""
        self._cellName = ""
        self._viewName = ""
        self._instanceName = ""
        self._netlistLine = ""
        # self._simViewName = None
        self._angle = 0.0
        self._drawings = list()
        self._labels = dict()  # dict of labels
        self._pins = dict()  # dict of pins
        self._netlistIgnore = False
        self.pinLocations = dict()  # pinName: pinRect
        self.pinNetMap = dict()  # pinName: netName
        self.pinNetTupleList = list()  # list of pinNetTuple
        for item in self.shapes:
            item.setFlag(QGraphicsItem.ItemIsSelectable, False)
            item.setFlag(QGraphicsItem.ItemStacksBehindParent, True)
            item.setParentItem(self)
            if type(item) is pin:
                self._pins[item.pinName] = item
            elif type(item) is label:
                self._labels[item.labelName] = item
            else:
                self._drawings.append(item)
        self.setFiltersChildEvents(True)
        self.setHandlesChildEvents(True)
        self.setFlag(QGraphicsItem.ItemContainsChildrenInShape, True)
        self.borderRect = self._drawings[0].sceneBoundingRect()
        if self._drawings[1:]:
            for draw in self._drawings[1:]:
                self.borderRect = self.borderRect.united(draw.sceneBoundingRect())
        self.dashLines = dict()

    def __repr__(self):
        return (f"symbolShape(name={self.cellName}, scene position= {self.scenePos()}, "
                f"pins = {self.pins}, labels = {self.labels}, ")

    def paint(self, painter, option, widget):
        if self.isSelected():
            painter.setPen(QPen(Qt.yellow, 2, Qt.DashLine))
            painter.drawRect(self.borderRect)
        if self.netlistIgnore:
            painter.setPen(QPen(Qt.red, 5, Qt.SolidLine))
            painter.drawLine(self.borderRect.bottomLeft(), self.borderRect.topRight())
            painter.drawLine(self.borderRect.topLeft(), self.borderRect.bottomRight())

    def boundingRect(self):
        return self.childrenBoundingRect()

    def sceneEvent(self, event):
        try:  # if net is being drawn, do not accept any event.
            if self.scene().drawWire:
                return False
            else:
                super().sceneEvent(event)
                return True
        except AttributeError:
            return False

    def itemChange(self, change, value):

        if self.scene() and change == QGraphicsItem.ItemPositionHasChanged:
            # item's position has changed
            # do something here
            for item in self.pinNetTupleList:
                if item.start == 1:
                    item.net.start = item.pin.mapToScene(item.pin.start)
                elif item.start == 0:
                    item.net.end = item.pin.mapToScene(item.pin.start)

        return super().itemChange(change, value)

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        self.pinNetTupleList = list()
        # create a named tuple to record whether the pin is located at the start or at
        # the end of the net.
        pins = [item for item in self.pins.values()]
        for pinItem in pins:
            for pinNet in self.scene().items(
                    pinItem.sceneBoundingRect().adjusted(-2, -2, 2, 2),
                    Qt.IntersectsItemShape):
                if isinstance(pinNet, net.schematicNet):
                    if pinItem.sceneBoundingRect().adjusted(-2, -2, 2, 2).contains(
                            pinNet.mapToScene(pinNet.start)):
                        self.pinNetTupleList.append(ddef.pinNetTuple(pinItem, pinNet, 1))
                    elif pinItem.sceneBoundingRect().adjusted(-2, -2, 2, 2).contains(
                            pinNet.mapToScene(pinNet.end)):
                        self.pinNetTupleList.append(ddef.pinNetTuple(pinItem, pinNet, 0))
        for item in self.pinNetTupleList:
            pinSceneLoc = item.pin.mapToScene(item.pin.start)
            if item.start == 1:
                item.net.start = pinSceneLoc
            else:
                item.net.end = pinSceneLoc
            item.net.pen = self.scene().otherPen
        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseReleaseEvent(event)

        for item in self.pinNetTupleList:
            lines = self.scene().addWires(item.net.start, self.scene().wirePen)
            self.scene().extendWires(lines, item.net.start, item.net.end)
            self.scene().pruneWires(lines, self.scene().wirePen)
            self.scene().removeItem(item.net)

    @property
    def libraryName(self):
        return self._libraryName

    @libraryName.setter
    def libraryName(self, value):
        self._libraryName = value

    @property
    def cellName(self):
        return self._cellName

    @cellName.setter
    def cellName(self, value: str):
        self._cellName = value

    @property
    def viewName(self):
        return self._viewName

    @viewName.setter
    def viewName(self, value: str):
        self._viewName = value

    @property
    def instanceName(self):
        return self._instanceName

    @instanceName.setter
    def instanceName(self, value: str):
        '''
        If instance name is changed and [@instName] label exists, change it too.
        '''
        self._instanceName = value
        if self.labels.get('instanceName', None):
            self.labels['instanceName'].labelValue = value
            self.labels['instanceName'].labelValueSet = True
            self.labels['instanceName'].update()

    @property
    def counter(self) -> int:
        return self._counter

    @counter.setter
    def counter(self, value: int):
        assert isinstance(value, int)
        self._counter = value

    @property
    def angle(self):
        return self._angle

    @angle.setter
    def angle(self, value: float):
        self.setRotation(value)
        self._angle = value

    @property
    def labels(self):
        return self._labels  # dictionary

    @property
    def pins(self):
        return self._pins

    @property
    def drawings(self):
        return self._drawings

    @property
    def netlistIgnore(self) -> bool:
        return self._netlistIgnore

    @netlistIgnore.setter
    def netlistIgnore(self, value: bool):
        self._netlistIgnore = value


class schematicPin(shape):
    """
    schematic pin class.
    """

    pinDirs = ["Input", "Output", "Inout"]
    pinTypes = ["Signal", "Ground", "Power", "Clock", "Digital", "Analog"]

    def __init__(self, start: QPoint, pen: QPen, pinName, pinDir, pinType,
                 gridTuple: tuple):
        super().__init__(pen, gridTuple)
        self._start = start
        self._pinName = pinName
        self._pinDir = pinDir
        self._pinType = pinType
        self._netTupleSet = set()

    def __repr__(self):
        return f"schematicPin({self._start}, {self._pen}, {self._pinName}, {self._pinDir}, {self._pinType})"

    def paint(self, painter, option, widget):

        painter.setPen(self._pen)
        painter.setBrush(self._pen.color())
        painter.setFont(QFont("Arial", 12))
        match self.pinDir:
            case "Input":
                painter.drawPolygon([QPoint(self._start.x() - 10, self._start.y() - 10),
                                     QPoint(self._start.x() + 10, self._start.y() - 10),
                                     QPoint(self._start.x() + 20, self._start.y()),
                                     QPoint(self._start.x() + 10, self._start.y() + 10),
                                     QPoint(self._start.x() - 10, self._start.y() + 10), ])
            case "Output":
                painter.drawPolygon([QPoint(self._start.x() - 20, self._start.y()),
                                     QPoint(self._start.x() - 10, self._start.y() - 10),
                                     QPoint(self._start.x() + 10, self._start.y() - 10),
                                     QPoint(self._start.x() + 10, self._start.y() + 10),
                                     QPoint(self._start.x() - 10, self._start.y() + 10), ])
            case "Inout":
                painter.drawPolygon([QPoint(self._start.x() - 20, self._start.y()),
                                     QPoint(self._start.x() - 10, self._start.y() - 10),
                                     QPoint(self._start.x() + 10, self._start.y() - 10),
                                     QPoint(self._start.x() + 20, self._start.y()),
                                     QPoint(self._start.x() + 10, self._start.y() + 10),
                                     QPoint(self._start.x() - 10, self._start.y() + 10), ])
        painter.drawText(self._start.x(), self._start.y() - 20, self.pinName)
        if self.isSelected():
            painter.setPen(QPen(Qt.yellow, 2, Qt.DashLine))
            # painter.setBrush(Qt.yellow)
            painter.drawRect(QRect.span(QPoint(self._start.x() - 10, self._start.y() - 10),
                                        QPoint(self._start.x() + 10,
                                               self._start.y() + 10), ))

    def boundingRect(self):
        return QRect(self._start.x() - 10, self._start.y() - 10, 30, 20).adjusted(-5, -10,
            5, 5)

    def sceneEvent(self, event):
        if self.scene().drawWire:
            return False
        else:
            super().sceneEvent(event)
            return True

    #
    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mousePressEvent(event)
        # create a named tuple to record whether the pin is located at the start or at
        # the end of the net.

        for pinNet in self.scene().items(self.sceneBoundingRect().adjusted(-2, -2, 2, 2),
                Qt.IntersectsItemShape):
            if isinstance(pinNet, net.schematicNet):
                if self.sceneBoundingRect().adjusted(-2, -2, 2, 2).contains(
                        pinNet.mapToScene(pinNet.start)):
                    self._netTupleSet.add(ddef.netTuple(pinNet, 1))
                elif self.sceneBoundingRect().adjusted(-2, -2, 2, 2).contains(
                        pinNet.mapToScene(pinNet.end)):
                    self._netTupleSet.add(ddef.netTuple(pinNet, 0))
        for item in self._netTupleSet:
            pinSceneLoc = self.mapToScene(self.start)
            if item.start == 1:
                item.net.start = pinSceneLoc
            else:
                item.net.end = pinSceneLoc
            item.net.pen = self.scene().otherPen
        super().mousePressEvent(event)

    #
    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseReleaseEvent(event)

        for item in self._netTupleSet:
            lines = self.scene().addWires(item.net.start, self.scene().wirePen)
            self.scene().extendWires(lines, item.net.start, item.net.end)
            self.scene().pruneWires(lines, self.scene().wirePen)
            self.scene().removeItem(item.net)
        self._netTupleSet = set()

    def itemChange(self, change, value):

        if self.scene() and change == QGraphicsItem.ItemPositionHasChanged:
            # item's position has changed
            # do something here
            for item in self._netTupleSet:
                if item.start == 1:
                    item.net.start = self.mapToScene(self.start)
                elif item.start == 0:
                    item.net.end = self.mapToScene(self.start)
        return super().itemChange(change, value)

    # def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
    #     super().mouseMoveEvent(event)
    #     self.setPos(event.scenePos() - event.buttonDownPos(Qt.LeftButton))

    def toSymbolPin(self, start: QPoint, pen: QPen, gridTuple: tuple):
        return pin(start, pen, self.pinName, self.pinDir, self.pinType, gridTuple)

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, start):
        self._start = start

    @property
    def pinName(self):
        return self._pinName

    @pinName.setter
    def pinName(self, pinName):
        if pinName != "":
            self._pinName = pinName

    @property
    def pinDir(self):
        return self._pinDir

    @pinDir.setter
    def pinDir(self, direction: str):
        if direction in self.pinDirs:
            self._pinDir = direction

    @property
    def pinType(self):
        return self._pinType

    @pinType.setter
    def pinType(self, pintype: str):
        if pintype in self.pinTypes:
            self._pinType = pintype
